//
// tclSound.cc
// 
// tcl extension to have sound files as tcl 'object-oriented' commands
// 
// dpwe@media.mit.edu 1994sep07
// uses my Sound.cc object.
//

#include <assert.h>

#include "Sound.h"
extern "C" {
#include "tcl.h"
#include "audIO.h"

char *programName = "ARGH! GROSS";

Tcl_HashTable soundTable;		// Convert sound names to ptrs

int Sound_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int soundCommand(ClientData clientData, Tcl_Interp *interp, 
		 int argc, char **argv);
int soundProc(ClientData clientData, Tcl_Interp *interp, 
	      int argc, char **argv);
int soundConfigure(Sound *soundPtr, Tcl_Interp *interp, 
	      char *cmdName, int argc, char **argv);
void soundDelete(ClientData clientData);
}

/* Usage:
 * sound sndObj -mode {read|write} -samplerate 44100 -chans 1 -format short
 * sndObj config -mode r -samplerate 44100 -chans 1 -format short
 * sndObj config -nframes
 * sndObj config -filename
 * 
 * sndObj open <filename>
 * sndObj read <frames>
 * sndObj write <frames>
 * sndObj seek <frame>
 * sndObj tell
 * sndObj eof
 * sndObj play
 */

int Sound_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tcl_CreateCommand(interp, "sound", soundCommand, (ClientData) NULL,
	    (void (*)(ClientData)) NULL);
    // initialize the hash table
    Tcl_InitHashTable(&soundTable, TCL_STRING_KEYS);
    return TCL_OK;
}

int soundCommand(ClientData clientData, Tcl_Interp *interp, 
		 int argc, char **argv)
{	/* tcl command to instance a new sound object */
    if(argc < 2) {
	Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
    // allocate a new object
    Sound *soundPtr = new Sound;
    // put it into the tcl command namespace
    Tcl_CreateCommand(interp, argv[1], soundProc, (ClientData)soundPtr, 
		      soundDelete);
    // .. and add it as a key in the hash table lookup
    int isnew;
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_CreateHashEntry(&soundTable, argv[1], &isnew);
    if(!isnew)  {		// re-using a name - must destroy old command
	// actually, command (and hence data structure) will have been 
	// deleted when we re-used the command name above, so it 
	// doesn't matter
    }
    Tcl_SetHashValue(entryPtr, soundPtr);

    // pass rest of command to configure
    if(soundConfigure(soundPtr, interp, argv[0], argc-2, argv+2) != TCL_OK)  {
	return TCL_ERROR;
    }
    interp->result = argv[1];
    return TCL_OK;
}

static void soundPlay(Sound *snd, long start, long len)
{
    Format	fmt   = snd->getFormat();
    double 	srate = snd->getSampleRate();
    int 	chans = snd->getChannels();
    long	desbuf= long(srate*0.2);	// 200 ms buffer
    long	totalframes = snd->getFrames();
    long 	remain, thistime, got = -1;
    int 	entryPos = snd->frameTell();
    Format 	playfmt = Sfmt_SHORT;

    AUSTRUCT *au = AUOpen(playfmt, srate, chans, desbuf, AIO_OUTPUT);
    if(au != NULL)  {
	snd->setFormat(playfmt);
	long myBsize = desbuf/2;
	short *myBuf = new short[chans*myBsize];
	if(start < 0 || start >= totalframes)	start = 0;
	if(len < 0 || start+len > totalframes)	len = totalframes - start;
	snd->frameSeek(start);
	remain = len;

	while(remain > 0 && got != 0)  {
	    thistime = MIN(myBsize, remain);
	    got = snd->readFrames(myBuf, thistime);
	    AUWrite(au, myBuf, got);
	    remain -= got;
	}
	AUClose(au, 0);
	delete [] myBuf;
	snd->setFormat(fmt);
	snd->frameSeek(entryPos);
    }
}

int soundProc(ClientData clientData, Tcl_Interp *interp, 
	      int argc, char **argv)
{	/* handle the commands supported by the soundObj */
    Sound *snd = (Sound *)clientData;
    int		result = TCL_OK;
    char	c;
    int		length;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if((c=='c') && strncmp(argv[1], "configure", length)==0)  {
	return soundConfigure(snd, interp, argv[0], argc-1, argv+1);
    }
    if( (c=='p') && strncmp(argv[1], "play", length)==0)  {
	int start=0, len=-1;
	if(argc>4)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " play [start [len]]\"", NULL);
	    return TCL_ERROR;
	}
	if(argc>3)  {
	    if(Tcl_GetInt(interp, argv[3], &len)!=TCL_OK)
		return TCL_ERROR;
	}
	if(argc>2)  {
	    if(Tcl_GetInt(interp, argv[2], &start)!=TCL_OK)
		return TCL_ERROR;
	}
	soundPlay(snd, start, len);
	return TCL_OK;
    } 
    if(argc > 3)  {	/* config, play only command with many args */
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?arg?\"", NULL);
	return TCL_ERROR;
    }
    if( (c=='o') && strncmp(argv[1], "open", length)==0)  {
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " open <filename>\"", NULL);
	    return TCL_ERROR;
	}
	if(snd->open(argv[2])!=0)  {
	    Tcl_AppendResult(interp, "error opening \"", argv[2], 
			     "\" as soundfile", NULL);
	    return TCL_ERROR;
	}
	interp->result = argv[2];
    } else if( (c=='r') && strncmp(argv[1], "read", length)==0)  {
	int frames;
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " read <framecount>\"", NULL);
	    return TCL_ERROR;
	}
	if(Tcl_GetInt(interp, argv[2], &frames)!=TCL_OK)
	    return TCL_ERROR;
	int chans = snd->getChannels();
	if(snd->getFormat()==Sfmt_SHORT)  {
	    short *s = new short[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = snd->readFrames(s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%d ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans-1; ++ch) {
			sprintf(str, "%d ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%d", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else if(snd->getFormat()==Sfmt_FLOAT)  {
	    float *s = new float[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = snd->readFrames(s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%.6f ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans-1; ++ch) {
			sprintf(str, "%.6f ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%.6f", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else if(snd->getFormat()==Sfmt_LONG)  {
	    long *s = new long[frames * chans];
	    int f, got, ch;
	    char str[32];
	    got = snd->readFrames(s, frames);
	    if(chans == 1) {
		for(f = 0; f<frames; ++f)  {
		    sprintf(str, "%ld ", s[f]);
		    Tcl_AppendResult(interp, str, NULL);
		}
	    } else {
		for(f = 0; f<frames; ++f)  {
		    Tcl_AppendResult(interp, "{", NULL);
		    for(ch = 0; ch<chans; ++ch) {
			sprintf(str, "%ld ", s[f*chans+ch]);
			Tcl_AppendResult(interp, str, NULL);
		    }
		    sprintf(str, "%ld", s[f*chans+ch]);
		    Tcl_AppendResult(interp, str, "} ", NULL);
		}
	    }
	} else {
	    Tcl_AppendResult(interp, 
			     "hopelessly confused about format in read");
	    return TCL_ERROR;
	}
//	sprintf(interp->result, "(read %d frames)", frames);
    } else if( (c=='w') && strncmp(argv[1], "write", length)==0)  {
	int frames;
	char **listels;
	int nlistels;

	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " write <datalist>\"", NULL);
	    return TCL_ERROR;
	}
	Tcl_SplitList(interp, argv[2], &nlistels, &listels);
	// process in here
	free(listels);
	sprintf(interp->result, "(wrote %d frames)", nlistels);
    } else if( (c=='s') && strncmp(argv[1], "seek", length)==0)  {
	int pos;
	if(argc!=3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " seek <pos>\"", NULL);
	    return TCL_ERROR;
	}
	if(Tcl_GetInt(interp, argv[2], &pos)!=TCL_OK)
	    return TCL_ERROR;
	pos = snd->frameSeek(pos);
	sprintf(interp->result, "%d", pos);
    } else if( (c=='t') && strncmp(argv[1], "tell", length)==0)  {
	int pos;
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " tell\"", NULL);
	    return TCL_ERROR;
	}
	pos = snd->frameTell();
	sprintf(interp->result, "%d", pos);
    } else if( (c=='e') && strncmp(argv[1], "eof", length)==0)  {
	int eof;
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " eof\"", NULL);
	    return TCL_ERROR;
	}
	eof = snd->frameFeof();
	sprintf(interp->result, "%d", eof);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1], 
		"\":  config/open/read/write/seek/tell/eof/play",
			 (char *) NULL);
	result = TCL_ERROR;
    }
    return result;
}

int soundConfigure(Sound *snd, Tcl_Interp *interp, 
		   char *cmdName, int argc, char **argv)
{	/* Handle arguments specifically to configure */
    int i = 0; 
    char c, *s;
    int len;

    while(++i < argc)  {
	if(argv[i][0] != '-')  {
	    Tcl_AppendResult(interp, "configure error: should be \"", 
			     cmdName, " config {-mode|-samplerate|-chans|-format|-nframes|-filename} ?val?", NULL);
	    return TCL_ERROR;
	}
	s = &(argv[i][1]);  c = *s;  len = strlen(s);
	if( (c=='m') && strncmp(s,"mode",len)==0)  {
	    char mc;
	    ++i;
	    if(i==argc)  {
		mc = snd->getMode();
	    } else {
		char *arg = argv[i];
		int larg = strlen(arg);
		char argi = *arg;
		if( (argi=='r') && strncmp(arg, "read", larg)==0)  {
		    mc = snd->setMode('r');
		} else if( (argi=='w') && strncmp(arg, "write", larg)==0)  {
		    mc = snd->setMode('w');
		} else {
		    Tcl_AppendResult(interp, "error: should be \"", 
				     cmdName, " mode {read|write}\"", NULL);
		    return TCL_ERROR;
		}
	    }
	    if(mc=='r')
		sprintf(interp->result, "read");
	    else
		sprintf(interp->result, "write");
	} else if( (c=='s' && strncmp(s,"samplerate",len)==0) ) {
	    double srate;
	    ++i;
	    if(i==argc)  {
		srate = snd->getSampleRate();
	    } else {
		if(Tcl_GetDouble(interp, argv[i], &srate)!=TCL_OK)
		    return TCL_ERROR;
		snd->setSampleRate(srate);
	    }
	    sprintf(interp->result, "%.2f", srate);
	} else if( (c=='c' && strncmp(s,"chans",len)==0) ) {
	    int chans;
	    ++i;
	    if(i==argc)  {
		chans = snd->getChannels();
	    } else {
		if(Tcl_GetInt(interp, argv[i], &chans)!=TCL_OK)
		    return TCL_ERROR;
		snd->setChannels(chans);
	    }
	    sprintf(interp->result, "%d", chans);
	} else if( (c=='f') && strncmp(s,"format",len)==0)  {
	    Format fmt;
	    ++i;
 	    if(i==argc)  {
		fmt = snd->getFormat();
	    } else {
		char *arg = argv[i];
		int larg = strlen(arg);
		char argi = *arg;
		if( (argi=='s') && strncmp(arg, "short", larg)==0)  {
		    fmt = Sfmt_SHORT;
		} else if( (argi=='l') && strncmp(arg, "long", larg)==0)  {
		    fmt = Sfmt_LONG;
		} else if( (argi=='f') && strncmp(arg, "float", larg)==0)  {
		    fmt = Sfmt_FLOAT;
		} else {
		    Tcl_AppendResult(interp, "error: should be \"", 
				     cmdName, " format {short|long|float}\"", NULL);
		    return TCL_ERROR;
		}
		snd->setFormat(fmt);
	    }
	    sprintf(interp->result, (fmt==Sfmt_SHORT)?"short":((fmt==Sfmt_LONG)?"long":((fmt==Sfmt_FLOAT)?"float":"unknown")));
	} else if( (c=='n' && strncmp(s,"nframes",len)==0) ) {
	    int frms;
	    ++i;
	    if(i==argc)  {
		frms = snd->getFrames();
		sprintf(interp->result, "%d", frms);
	    } else {
		Tcl_AppendResult(interp, "wrong # args: should be \"", 
				cmdName, " nframes", NULL);
		return TCL_ERROR;
	    }
	} else if( (c=='f' && strncmp(s,"filename",len)==0) ) {
	    ++i;
	    if(i==argc)  {
		sprintf(interp->result, "%s", snd->getFilename());
	    } else {
		Tcl_AppendResult(interp, "wrong # args: should be \"", 
				cmdName, " filename", NULL);
		return TCL_ERROR;
	    }
	} else {
	    Tcl_AppendResult(interp, "unrecognized: should be \"", 
			     cmdName, "config {-mode|-samplerate|-chans|-format|-nframes|-filename} ?val?", NULL);
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

void soundDelete(ClientData clientData)
{	/* a particular soundObj is being deleted */
    Sound *snd = (Sound *)clientData;
    delete snd;
}
    
